@@ -32,20 +32,17 @@ import java.util.HashSet;
 import java.util.Locale;
 import java.util.Map;
 import java.util.Set;
-import java.util.TreeMap;
 import java.util.function.Supplier;
 
 import javax.servlet.ServletException;
 import javax.servlet.ServletOutputStream;
 import javax.servlet.SessionTrackingMode;
+import javax.servlet.http.Cookie;
 import javax.servlet.http.HttpServletResponse;
 import javax.servlet.http.HttpSession;
 
 import io.undertow.UndertowLogger;
-import io.undertow.server.Connectors;
 import io.undertow.server.HttpServerExchange;
-import io.undertow.server.ResponseCommitListener;
-import io.undertow.server.handlers.Cookie;
 import io.undertow.server.protocol.http.HttpAttachments;
 import io.undertow.servlet.UndertowServletMessages;
 import io.undertow.servlet.handlers.ServletRequestContext;
@@ -58,12 +55,16 @@ import io.undertow.util.HttpString;
 import io.undertow.util.Protocols;
 import io.undertow.util.RedirectBuilder;
 import io.undertow.util.StatusCodes;
+import org.xnio.Bits;
 
 import static io.undertow.util.URLUtils.isAbsoluteUrl;
 
-
 /**
+ * Implementation of {@code HttpServletResponse}.
+ *
  * @author Stuart Douglas
+ * @author <a href="mailto:ropalka@redhat.com">Richard Opalka</a>
+ * @author Flavia Rainone
  */
 public final class HttpServletResponseImpl implements HttpServletResponse {
 
@@ -76,20 +77,25 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     private PrintWriter writer;
     private Integer bufferSize;
     private long contentLength = -1;
-    private boolean insideInclude = false;
+
+    private int flags = 0;
+    // response is inside include
+    private static final int INSIDE_INCLUDE_FLAG          = 1 << 0x00;
+    // response is done
+    private static final int RESPONSE_DONE_FLAG           = 1 << 0x01;
+    // ignored flush has been performed
+    private static final int IGNORED_FLUSH_PERFORMED_FLAG = 1 << 0x02;
+    // prevents anything to be added to response
+    private static final int TREAT_AS_COMMITTED_FLAG      = 1 << 0x03;
+    // indicates that we are closing the response and no further content should be written
+    private static final int CONTENT_FULLY_WRITTEN_FLAG   = 1 << 0x04;
+    //if a content type has been set either implicitly or implicitly
+    private static final int CHARSET_SET_FLAG             = 1 << 0x05;
+
     private Locale locale;
-    private boolean responseDone = false;
-
-    private boolean ignoredFlushPerformed = false;
-
-    private boolean treatAsCommitted = false;
-
-    private boolean charsetSet = false; //if a content type has been set either implicitly or implicitly
     private String contentType;
     private String charset;
     private Supplier<Map<String, String>> trailerSupplier;
-    // a map of cookie name -> a map of (cookie path + cookie domain) -> cookie
-    private Map<String, Map<String, Cookie>> duplicateCookies;
 
     public HttpServletResponseImpl(final HttpServerExchange exchange, final ServletContextImpl servletContext) {
         this.exchange = exchange;
@@ -102,46 +108,14 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     @Override
-    public void addCookie(final javax.servlet.http.Cookie cookie) {
-        if (insideInclude) {
+    public void addCookie(final Cookie newCookie) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG)) {
             return;
         }
-        final ServletCookieAdaptor servletCookieAdaptor = new ServletCookieAdaptor(cookie);
-        if (cookie.getVersion() == 0) {
+        final ServletCookieAdaptor servletCookieAdaptor = new ServletCookieAdaptor(newCookie);
+        if (newCookie.getVersion() == 0) {
             servletCookieAdaptor.setVersion(servletContext.getDeployment().getDeploymentInfo().getDefaultCookieVersion());
         }
-        // test for duplicate entry
-        if (exchange.getResponseCookies().containsKey(servletCookieAdaptor.getName())) {
-            final String cookieName = servletCookieAdaptor.getName();
-            final String path = servletCookieAdaptor.getPath();
-            final String domain = servletCookieAdaptor.getDomain();
-            final Cookie otherCookie = exchange.getResponseCookies().get(cookieName);
-            final String otherCookiePath = otherCookie.getPath();
-            final String otherCookieDomain = otherCookie.getDomain();
-            // if both cookies have same path, name, and domain, overwrite previous cookie
-            if ((path == otherCookiePath || (path != null && path.equals(otherCookiePath))) &&
-                    (domain == otherCookieDomain || (domain != null && domain.equals(otherCookieDomain)))) {
-                exchange.setResponseCookie(servletCookieAdaptor);
-            }
-            // else, create a duplicate cookie entry
-            else {
-                final Map<String, Cookie> cookiesByPathPlusDomain;
-                if (duplicateCookies == null) {
-                    duplicateCookies = new TreeMap<>();
-                    exchange.addResponseCommitListener(
-                            new DuplicateCookieCommitListener());
-                }
-                if (duplicateCookies.containsKey(cookieName)) {
-                    cookiesByPathPlusDomain = duplicateCookies.get(cookieName);
-                } else {
-                    cookiesByPathPlusDomain = new TreeMap<>();
-                    duplicateCookies.put(cookieName, cookiesByPathPlusDomain);
-                }
-                String pathPlusDomain = (otherCookiePath == null ? "null" : otherCookiePath) +
-                        (otherCookieDomain == null? "null" : otherCookieDomain);
-                cookiesByPathPlusDomain.put(pathPlusDomain, otherCookie);
-            }
-        }
         exchange.setResponseCookie(servletCookieAdaptor);
     }
 
@@ -162,7 +136,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
 
     @Override
     public void sendError(final int sc, final String msg) throws IOException {
-        if(insideInclude) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG)) {
             //not 100% sure this is the correct action
             return;
         }
@@ -181,7 +155,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         exchange.setStatusCode(sc);
         if(src.isRunningInsideHandler()) {
             //all we do is set the error on the context, we handle it when the request is returned
-            treatAsCommitted = true;
+            flags |= TREAT_AS_COMMITTED_FLAG;
             src.setError(sc, msg);
         } else {
             //if the src is null there is no outer handler, as we are in an asnc request
@@ -193,7 +167,8 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         writer = null;
         responseState = ResponseState.NONE;
         resetBuffer();
-        treatAsCommitted = false;
+        exchange.getResponseHeaders().remove(Headers.CONTENT_LENGTH);
+        flags &= ~TREAT_AS_COMMITTED_FLAG;
         final String location = servletContext.getDeployment().getErrorPages().getErrorLocation(sc);
         if (location != null) {
             RequestDispatcherImpl requestDispatcher = new RequestDispatcherImpl(location, servletContext);
@@ -227,6 +202,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
             throw UndertowServletMessages.MESSAGES.responseAlreadyCommited();
         }
         resetBuffer();
+        exchange.getResponseHeaders().remove(Headers.CONTENT_LENGTH);
         setStatus(StatusCodes.FOUND);
         String realPath;
         if (isAbsoluteUrl(location)) {//absolute url
@@ -240,8 +216,17 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
                 int lastSlash = current.lastIndexOf("/");
                 if (lastSlash != -1) {
                     current = current.substring(0, lastSlash + 1);
+                } else {
+                    current = "";
                 }
-                realPath = CanonicalPathUtils.canonicalize(servletContext.getContextPath() + current + location);
+                String precanonLocation = location;
+                String query = "";
+                int firstQuestionMark = location.indexOf("?");
+                if (firstQuestionMark >= 0) {
+                    precanonLocation = location.substring(0, firstQuestionMark);
+                    query = location.substring(firstQuestionMark);
+                }
+                realPath = CanonicalPathUtils.canonicalize(servletContext.getContextPath() + current + precanonLocation) + query;
             }
             String loc = exchange.getRequestScheme() + "://" + exchange.getHostAndPort() + realPath;
             exchange.getResponseHeaders().put(Headers.LOCATION, loc);
@@ -272,7 +257,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         if(name == null) {
             throw UndertowServletMessages.MESSAGES.headerNameWasNull();
         }
-        if (insideInclude || ignoredFlushPerformed) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG | IGNORED_FLUSH_PERFORMED_FLAG)) {
             return;
         }
         if(name.equals(Headers.CONTENT_TYPE)) {
@@ -294,7 +279,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         if(name == null) {
             throw UndertowServletMessages.MESSAGES.headerNameWasNull();
         }
-        if (insideInclude || ignoredFlushPerformed || treatAsCommitted) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG | IGNORED_FLUSH_PERFORMED_FLAG | TREAT_AS_COMMITTED_FLAG | CONTENT_FULLY_WRITTEN_FLAG)) {
             return;
         }
         if(name.equals(Headers.CONTENT_TYPE) && !exchange.getResponseHeaders().contains(Headers.CONTENT_TYPE)) {
@@ -316,7 +301,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
 
     @Override
     public void setStatus(final int sc) {
-        if (insideInclude || treatAsCommitted) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG | TREAT_AS_COMMITTED_FLAG)) {
             return;
         }
         if (responseStarted()) {
@@ -328,7 +313,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     @Override
     public void setStatus(final int sc, final String sm) {
         setStatus(sc);
-        if(!insideInclude && servletContext.getDeployment().getDeploymentInfo().isSendCustomReasonPhraseOnError()) {
+        if(!Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG) && servletContext.getDeployment().getDeploymentInfo().isSendCustomReasonPhraseOnError()) {
             exchange.setReasonPhrase(sm);
         }
     }
@@ -382,7 +367,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     @Override
     public String getContentType() {
         if (contentType != null) {
-            if (charsetSet) {
+            if (Bits.anyAreSet(flags, CHARSET_SET_FLAG)) {
                 return contentType + ";charset=" + getCharacterEncoding();
             } else {
                 return contentType;
@@ -404,7 +389,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     @Override
     public PrintWriter getWriter() throws IOException {
         if (writer == null) {
-            if (!charsetSet) {
+            if (!Bits.anyAreSet(flags, CHARSET_SET_FLAG)) {
                 //servet 5.5
                 setCharacterEncoding(getCharacterEncoding());
             }
@@ -431,10 +416,14 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
 
     @Override
     public void setCharacterEncoding(final String charset) {
-        if (insideInclude || responseStarted() || writer != null || isCommitted()) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG) || responseStarted() || writer != null || isCommitted()) {
             return;
         }
-        charsetSet = charset != null;
+        if (charset != null) {
+            flags |= CHARSET_SET_FLAG;
+        } else {
+            flags &= ~CHARSET_SET_FLAG;
+        }
         this.charset = charset;
         if (contentType != null) {
             exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, getContentType());
@@ -448,7 +437,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
 
     @Override
     public void setContentLengthLong(final long len) {
-        if (insideInclude || responseStarted()) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG) || responseStarted()) {
             return;
         }
         if(len >= 0) {
@@ -460,20 +449,24 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     boolean isIgnoredFlushPerformed() {
-        return ignoredFlushPerformed;
+        return Bits.anyAreSet(flags, IGNORED_FLUSH_PERFORMED_FLAG);
     }
 
     void setIgnoredFlushPerformed(boolean ignoredFlushPerformed) {
-        this.ignoredFlushPerformed = ignoredFlushPerformed;
+        if (ignoredFlushPerformed) {
+            flags |= IGNORED_FLUSH_PERFORMED_FLAG;
+        } else {
+            flags &= ~IGNORED_FLUSH_PERFORMED_FLAG;
+        }
     }
 
     private boolean responseStarted() {
-        return exchange.isResponseStarted() || ignoredFlushPerformed || treatAsCommitted;
+        return exchange.isResponseStarted() || Bits.anyAreSet(flags, IGNORED_FLUSH_PERFORMED_FLAG | TREAT_AS_COMMITTED_FLAG);
     }
 
     @Override
     public void setContentType(final String type) {
-        if (type == null || insideInclude || responseStarted()) {
+        if (type == null || Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG) || responseStarted()) {
             return;
         }
         ContentTypeInfo ct = servletContext.parseContentType(type);
@@ -481,10 +474,10 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         boolean useCharset = false;
         if(ct.getCharset() != null && writer == null && !isCommitted()) {
             charset = ct.getCharset();
-            charsetSet = true;
+            flags |= CHARSET_SET_FLAG;
             useCharset = true;
         }
-        if(useCharset || !charsetSet) {
+        if(useCharset || !Bits.anyAreSet(flags, CHARSET_SET_FLAG)) {
             exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, ct.getHeader());
         } else if(ct.getCharset() == null) {
             exchange.getResponseHeaders().put(Headers.CONTENT_TYPE, ct.getHeader() + "; charset=" + charset);
@@ -522,7 +515,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     public void closeStreamAndWriter() throws IOException {
-        if(treatAsCommitted) {
+        if(Bits.anyAreSet(flags, TREAT_AS_COMMITTED_FLAG)) {
             return;
         }
         if (writer != null) {
@@ -575,17 +568,17 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         responseState = ResponseState.NONE;
         exchange.getResponseHeaders().clear();
         exchange.setStatusCode(StatusCodes.OK);
-        treatAsCommitted = false;
+        flags &= ~(TREAT_AS_COMMITTED_FLAG | CONTENT_FULLY_WRITTEN_FLAG);
     }
 
     @Override
     public void setLocale(final Locale loc) {
-        if (insideInclude || responseStarted()) {
+        if (Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG) || responseStarted()) {
             return;
         }
         this.locale = loc;
         exchange.getResponseHeaders().put(Headers.CONTENT_LANGUAGE, loc.getLanguage() + "-" + loc.getCountry());
-        if (!charsetSet && writer == null) {
+        if (!Bits.anyAreSet(flags, CHARSET_SET_FLAG) && writer == null) {
             final Map<String, String> localeCharsetMapping = servletContext.getDeployment().getDeploymentInfo().getLocaleCharsetMapping();
             // Match full language_country_variant first, then language_country,
             // then language only
@@ -616,10 +609,10 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     public void responseDone() {
-        if (responseDone || treatAsCommitted) {
+        if (Bits.anyAreSet(flags, RESPONSE_DONE_FLAG | TREAT_AS_COMMITTED_FLAG)) {
             return;
         }
-        responseDone = true;
+        flags |= RESPONSE_DONE_FLAG;
         try {
             closeStreamAndWriter();
         } catch (IOException e) {
@@ -630,11 +623,15 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     public boolean isInsideInclude() {
-        return insideInclude;
+        return Bits.anyAreSet(flags, INSIDE_INCLUDE_FLAG);
     }
 
     public void setInsideInclude(final boolean insideInclude) {
-        this.insideInclude = insideInclude;
+        if (insideInclude) {
+            this.flags |= INSIDE_INCLUDE_FLAG;
+        } else {
+            this.flags &= ~INSIDE_INCLUDE_FLAG;
+        }
     }
 
     public void setServletContext(final ServletContextImpl servletContext) {
@@ -823,7 +820,7 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
     }
 
     public boolean isTreatAsCommitted() {
-        return treatAsCommitted;
+        return Bits.anyAreSet(flags, TREAT_AS_COMMITTED_FLAG);
     }
 
     @Override
@@ -854,16 +851,11 @@ public final class HttpServletResponseImpl implements HttpServletResponse {
         return trailerSupplier;
     }
 
-    private class DuplicateCookieCommitListener implements
-            ResponseCommitListener {
-
-        @Override
-        public void beforeCommit(HttpServerExchange exchange) {
-            for (Map.Entry<String, Map<String, Cookie>> duplicateCookiesEntry : duplicateCookies.entrySet()) {
-                for (Map.Entry<String, Cookie> cookiesByPathEntry : duplicateCookiesEntry.getValue().entrySet()) {
-                    Connectors.addCookie(exchange, cookiesByPathEntry.getValue());
-                }
-            }
-        }
+    /**
+     * Marks this response as closed for writing extra bytes, including the addition of headers.
+     */
+    void setContentFullyWritten() {
+        this.flags |= CONTENT_FULLY_WRITTEN_FLAG;
     }
+
 }
